/*
 * Copyright: Vitaly Puzrin
 * Author: Sergey Batishchev <snb2003@rambler.ru>
 *
 * Written for fontello.com project.
 */

import { encode, decode } from './lib/ucs2.js'

import SvgPath from 'svgpath'

import { load, cubicToQuad, toSfntCoutours } from './lib/svg.js'
import { Font, Glyph, Contour, Point } from './lib/sfnt.js'
import { generateTTF } from './lib/ttf.js'

let VERSION_RE = /^(Version )?(\d+[.]\d+)$/i

export function svg2ttf(svgString, options) {
  let font = new Font()
  let svgFont = load(svgString)

  options = options || {}

  font.id = options.id || svgFont.id
  font.familyName = options.familyname || svgFont.familyName || svgFont.id
  font.copyright = options.copyright || svgFont.metadata
  font.description = options.description || ''
  font.url = options.url || ''
  font.sfntNames.push({
    id: 2,
    value: options.subfamilyname || svgFont.subfamilyName || 'Regular'
  }) // subfamily name
  font.sfntNames.push({ id: 4, value: options.fullname || svgFont.id }) // full name

  let versionString = options.version || 'Version 1.0'

  if (!VERSION_RE.test(versionString)) {
    throw new Error(
      'svg2ttf: invalid option, version - "' + options.version + '"'
    )
  }

  versionString = 'Version ' + versionString.match(VERSION_RE)[2]
  font.sfntNames.push({ id: 5, value: versionString }) // version ID for TTF name table
  font.sfntNames.push({
    id: 6,
    value: (options.fullname || svgFont.id)
      .replace(/[\s\(\)\[\]<>%\/]/g, '')
      .substr(0, 62)
  }) // Postscript name for the font, required for OSX Font Book

  if (typeof options.ts !== 'undefined') {
    font.createdDate = font.modifiedDate = new Date(
      parseInt(options.ts, 10) * 1000
    )
  }

  // Try to fill font metrics or guess defaults
  //
  font.unitsPerEm = svgFont.unitsPerEm || 1000
  font.horizOriginX = svgFont.horizOriginX || 0
  font.horizOriginY = svgFont.horizOriginY || 0
  font.vertOriginX = svgFont.vertOriginX || 0
  font.vertOriginY = svgFont.vertOriginY || 0
  font.width = svgFont.width || svgFont.unitsPerEm
  font.height = svgFont.height || svgFont.unitsPerEm
  font.descent = !isNaN(svgFont.descent) ? svgFont.descent : -font.vertOriginY
  font.ascent = svgFont.ascent || font.unitsPerEm - font.vertOriginY
  // Values for font substitution. We're mostly working with icon fonts, so they aren't expected to be substituted.
  // https://docs.microsoft.com/en-us/typography/opentype/spec/os2#sxheight
  font.capHeight = svgFont.capHeight || 0 // 0 is a valid value if "H" glyph doesn't exist
  font.xHeight = svgFont.xHeight || 0 // 0 is a valid value if "x" glyph doesn't exist

  if (typeof svgFont.weightClass !== 'undefined') {
    let wght = parseInt(svgFont.weightClass, 10)

    if (!isNaN(wght)) font.weightClass = wght
    else {
      // Unknown names are silently ignored
      if (svgFont.weightClass === 'normal') font.weightClass = 400
      if (svgFont.weightClass === 'bold') font.weightClass = 700
    }
  }

  if (typeof svgFont.underlinePosition !== 'undefined') {
    font.underlinePosition = svgFont.underlinePosition
  }
  if (typeof svgFont.underlineThickness !== 'undefined') {
    font.underlineThickness = svgFont.underlineThickness
  }

  let glyphs = font.glyphs
  let codePoints = font.codePoints
  let ligatures = font.ligatures

  function addCodePoint(codePoint, glyph) {
    if (codePoints[codePoint]) {
      // Ignore code points already defined
      return false
    }
    codePoints[codePoint] = glyph
    return true
  }

  // add SVG glyphs to SFNT font
  svgFont.glyphs.forEach(function (svgGlyph) {
    let glyph = new Glyph()

    glyph.name = svgGlyph.name
    glyph.codes = svgGlyph.ligatureCodes || svgGlyph.unicode // needed for nice validator error output
    glyph.d = svgGlyph.d
    glyph.height = !isNaN(svgGlyph.height) ? svgGlyph.height : font.height
    glyph.width = !isNaN(svgGlyph.width) ? svgGlyph.width : font.width
    glyphs.push(glyph)

    svgGlyph.sfntGlyph = glyph

    svgGlyph.unicode.forEach(function (codePoint) {
      addCodePoint(codePoint, glyph)
    })
  })

  let missingGlyph

  // add missing glyph to SFNT font
  // also, check missing glyph existance and single instance
  if (svgFont.missingGlyph) {
    missingGlyph = new Glyph()
    missingGlyph.d = svgFont.missingGlyph.d
    missingGlyph.height = !isNaN(svgFont.missingGlyph.height)
      ? svgFont.missingGlyph.height
      : font.height
    missingGlyph.width = !isNaN(svgFont.missingGlyph.width)
      ? svgFont.missingGlyph.width
      : font.width
  } else {
    missingGlyph = glyphs.find(function (glyph) {
      return glyph.name === '.notdef'
    })
  }
  if (!missingGlyph) {
    // no missing glyph and .notdef glyph, we need to create missing glyph
    missingGlyph = new Glyph()
  }

  // Create glyphs for all characters used in ligatures
  svgFont.ligatures.forEach(function (svgLigature) {
    let ligature = {
      ligature: svgLigature.ligature,
      unicode: svgLigature.unicode,
      glyph: svgLigature.glyph.sfntGlyph
    }

    ligature.unicode.forEach(function (charPoint) {
      // We need to have a distinct glyph for each code point so we can reference it in GSUB
      let glyph = new Glyph()
      let added = addCodePoint(charPoint, glyph)

      if (added) {
        glyph.name = encode([charPoint])
        glyphs.push(glyph)
      }
    })
    ligatures.push(ligature)
  })

  // Missing Glyph needs to have index 0
  if (glyphs.indexOf(missingGlyph) !== -1) {
    glyphs.splice(glyphs.indexOf(missingGlyph), 1)
  }
  glyphs.unshift(missingGlyph)

  let nextID = 0

  //add IDs
  glyphs.forEach(function (glyph) {
    glyph.id = nextID
    nextID++

    // Calculate accuracy for cubicToQuad transformation
    // For glyphs with height and width smaller than 500 use relative 0.06% accuracy,
    // for larger glyphs use fixed accuracy 0.3.
    let glyphSize = Math.max(glyph.width, glyph.height)
    let accuracy = glyphSize > 500 ? 0.3 : glyphSize * 0.0006

    //SVG transformations
    let svgPath = new SvgPath(glyph.d)
      .abs()
      .unshort()
      .unarc()
      .iterate(function (segment, index, x, y) {
        return cubicToQuad(segment, index, x, y, accuracy)
      })
    let sfntContours = toSfntCoutours(svgPath)

    // Add contours to SFNT font
    glyph.contours = sfntContours.map(function (sfntContour) {
      let contour = new Contour()

      contour.points = sfntContour.map(function (sfntPoint) {
        let point = new Point()

        point.x = sfntPoint.x
        point.y = sfntPoint.y
        point.onCurve = sfntPoint.onCurve
        return point
      })

      return contour
    })
  })

  let ttf = generateTTF(font)

  return ttf
}
