/**
 * {}
 * @author yutent<yutent.io@gmail.com>
 * @date 2024/02/04 10:50:22
 */

import { decode } from './ucs2.js'

import cubic2quad from 'cubic2quad'
import svgpath from 'svgpath'
import { DOMParser } from '@xmldom/xmldom'

function getGlyph(glyphElem, fontInfo) {
  let glyph = {}

  if (glyphElem.hasAttribute('d')) {
    glyph.d = glyphElem.getAttribute('d').trim()
  } else {
    // try nested <path>
    let pathElem = glyphElem.getElementsByTagName('path')[0]

    if (pathElem.hasAttribute('d')) {
      // <path> has reversed Y axis
      glyph.d = svgpath(pathElem.getAttribute('d'))
        .scale(1, -1)
        .translate(0, fontInfo.ascent)
        .toString()
    } else {
      throw new Error("Can't find 'd' attribute of <glyph> tag.")
    }
  }

  glyph.unicode = []

  if (glyphElem.getAttribute('unicode')) {
    glyph.character = glyphElem.getAttribute('unicode')
    let unicode = decode(glyph.character)

    // If more than one code point is involved, the glyph is a ligature glyph
    if (unicode.length > 1) {
      glyph.ligature = glyph.character
      glyph.ligatureCodes = unicode
    } else {
      glyph.unicode.push(unicode[0])
    }
  }

  glyph.name = glyphElem.getAttribute('glyph-name')

  if (glyphElem.getAttribute('horiz-adv-x')) {
    glyph.width = parseInt(glyphElem.getAttribute('horiz-adv-x'), 10)
  }

  return glyph
}

function deduplicateGlyps(glyphs, ligatures) {
  let result = []

  glyphs.forEach(function (glyph) {
    let canonical = result.find(
      it => it.width === glyph.width && it.d === glyph.d
    )

    if (canonical) {
      canonical.unicode = canonical.unicode.concat(glyph.unicode)
      glyph.canonical = canonical
    } else {
      result.push(glyph)
    }
  })

  // Update ligatures to point to the canonical version
  ligatures.forEach(function (ligature) {
    while (ligature.glyph.canonical) {
      ligature.glyph = ligature.glyph.canonical
    }
  })

  return result
}

export function load(str) {
  let attrs

  let doc = new DOMParser().parseFromString(str, 'application/xml')

  let metadata, fontElem, fontFaceElem

  metadata = doc.getElementsByTagName('metadata')[0]
  fontElem = doc.getElementsByTagName('font')[0]

  if (!fontElem) {
    throw new Error(
      "Can't find <font> tag. Make sure you SVG file is font, not image."
    )
  }

  fontFaceElem = fontElem.getElementsByTagName('font-face')[0]

  let familyName = fontFaceElem.getAttribute('font-family') || 'fontello'
  let subfamilyName = fontFaceElem.getAttribute('font-style') || 'Regular'
  let id =
    fontElem.getAttribute('id') ||
    (familyName + '-' + subfamilyName)
      .replace(/[\s\(\)\[\]<>%\/]/g, '')
      .substr(0, 62)

  let font = {
    id: id,
    familyName: familyName,
    subfamilyName: subfamilyName,
    stretch: fontFaceElem.getAttribute('font-stretch') || 'normal'
  }

  // Doesn't work with complex content like <strong>Copyright:></strong><em>Fontello</em>
  if (metadata && metadata.textContent) {
    font.metadata = metadata.textContent
  }

  // Get <font> numeric attributes
  attrs = {
    width: 'horiz-adv-x',
    //height:       'vert-adv-y',
    horizOriginX: 'horiz-origin-x',
    horizOriginY: 'horiz-origin-y',
    vertOriginX: 'vert-origin-x',
    vertOriginY: 'vert-origin-y'
  }
  attrs.forEach(function (val, key) {
    if (fontElem.hasAttribute(val)) {
      font[key] = parseInt(fontElem.getAttribute(val), 10)
    }
  })

  // Get <font-face> numeric attributes
  attrs = {
    ascent: 'ascent',
    descent: 'descent',
    unitsPerEm: 'units-per-em',
    capHeight: 'cap-height',
    xHeight: 'x-height',
    underlineThickness: 'underline-thickness',
    underlinePosition: 'underline-position'
  }
  attrs.forEach(function (val, key) {
    if (fontFaceElem.hasAttribute(val)) {
      font[key] = parseInt(fontFaceElem.getAttribute(val), 10)
    }
  })

  if (fontFaceElem.hasAttribute('font-weight')) {
    font.weightClass = fontFaceElem.getAttribute('font-weight')
  }

  let missingGlyphElem = fontElem.getElementsByTagName('missing-glyph')[0]

  if (missingGlyphElem) {
    font.missingGlyph = {}
    font.missingGlyph.d = missingGlyphElem.getAttribute('d') || ''

    if (missingGlyphElem.getAttribute('horiz-adv-x')) {
      font.missingGlyph.width = parseInt(
        missingGlyphElem.getAttribute('horiz-adv-x'),
        10
      )
    }
  }

  let glyphs = []
  let ligatures = []

  fontElem.getElementsByTagName('glyph').forEach(function (glyphElem) {
    let glyph = getGlyph(glyphElem, font)

    if (glyph.ligature) {
      ligatures.push({
        ligature: glyph.ligature,
        unicode: glyph.ligatureCodes,
        glyph: glyph
      })
    }

    glyphs.push(glyph)
  })

  glyphs = deduplicateGlyps(glyphs, ligatures)

  font.glyphs = glyphs
  font.ligatures = ligatures

  return font
}

export function cubicToQuad(segment, index, x, y, accuracy) {
  if (segment[0] === 'C') {
    let quadCurves = cubic2quad(
      x,
      y,
      segment[1],
      segment[2],
      segment[3],
      segment[4],
      segment[5],
      segment[6],
      accuracy
    )

    let res = []

    for (let i = 2; i < quadCurves.length; i += 4) {
      res.push([
        'Q',
        quadCurves[i],
        quadCurves[i + 1],
        quadCurves[i + 2],
        quadCurves[i + 3]
      ])
    }
    return res
  }
}

// Converts svg points to contours.  All points must be converted
// to relative ones, smooth curves must be converted to generic ones
// before this conversion.
//
export function toSfntCoutours(svgPath) {
  let resContours = []
  let resContour = []

  svgPath.iterate(function (segment, index, x, y) {
    //start new contour
    if (index === 0 || segment[0] === 'M') {
      resContour = []
      resContours.push(resContour)
    }

    let name = segment[0]

    if (name === 'Q') {
      //add control point of quad spline, it is not on curve
      resContour.push({ x: segment[1], y: segment[2], onCurve: false })
    }

    // add on-curve point
    if (name === 'H') {
      // vertical line has Y coordinate only, X remains the same
      resContour.push({ x: segment[1], y: y, onCurve: true })
    } else if (name === 'V') {
      // horizontal line has X coordinate only, Y remains the same
      resContour.push({ x: x, y: segment[1], onCurve: true })
    } else if (name !== 'Z') {
      // for all commands (except H and V) X and Y are placed in the end of the segment
      resContour.push({
        x: segment[segment.length - 2],
        y: segment[segment.length - 1],
        onCurve: true
      })
    }
  })
  return resContours
}
